from YTools.universe.extra_type import Struct
from .utils import find_or_new , make_obj_list

def make_get(name):	
	'''Object类对象的定制属性的get函数，在访问对应属性的时候查找数据库对象
	'''

	def gets(self):
		return getattr(self.sql_obj , self.name_map[name] )
	return gets

def make_set(name):
	'''Object类对象的定制属性的set函数，在访问对应属性的时候查找数据库对象
	'''
	def sets(self , val):
		setattr( self.sql_obj , self.name_map[name] , val )
		self.sql_obj.save()
	return sets

class Object(Struct):

	def __init__(self , sql_class , from_obj , force_new = False , **kwargs):
		'''
		
		参数：
			sql_class：对应的 sql 类
			from_obj：用于初始化的 sql 对象。如果为 None 则根据 kwargs 新建或者查找
		'''

		super().__init__()

		self.sql_obj = from_obj
		if force_new:
			self.sql_obj = sql_class(from_obj , **kwargs)
		else:
			if from_obj is None:
				self.sql_obj = find_or_new(sql_class , **kwargs)

		self.sql_obj.save()

	def set_name_map(self , **kwargs):
		self.name_map = kwargs
		for name in kwargs:
			self._set_property(
				name , 
				make_get(name) , 
				make_set(name) , 
			)

	def save(self):
		self.sql_obj.save()


# Generated by Django 3.0.3 on 2020-07-29 11:07

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('experiment', '0022_auto_20200728_2051'),
    ]

    operations = [
        migrations.AddField(
            model_name='groupconfig',
            name='fixed_left',
            field=models.TextField(default='', null=True),
        ),
        migrations.AddField(
            model_name='groupconfig',
            name='fixed_right',
            field=models.TextField(default='', null=True),
        ),
    ]



import random
from .constants import cookie_key , debugging

passwords = []
my_seed = str(random.randint(2333 , 23333333))

if debugging:
	my_seed = str(233333)

def check_passwords():
	if len(passwords) == 0:
		if debugging:
			passwords.append("2333")
		else:
			passwords.append(input("input a possword:"))
	return passwords

def check_permission(request):
	check_passwords()
	return ask_cookie_value(request , cookie_key) == my_seed

def give_permission(request , response):
	'''给request的源发许可'''
	check_passwords()
	return set_cookie_value(response , key = cookie_key , value = my_seed)

def ask_cookie_value(request , key):
	'''输入一个request，返回这个request对应的cookie值

	:param request: 一个用户请求
	:param str key: cookie名 
	:return str: 这个请求的名为key的cookie的值，不存在则返回None
	'''
	return request.COOKIES.get(key)

def set_cookie_value(response , key , value = None):
	'''给用户设置cookie值

	:param response: 一个网站响应
	:param str key: cookie名 
	:param str value: cookie值
	:return response: 返回输入的response
	'''
	response.set_cookie(
		key = key ,
		value = value ,
		max_age = 60 * 60 * 24 * 365 * 100 , #每过一百年就得重新获取一次
		#domain =  domain,
		#path = "/" , 
	)
	return response

def del_cookie(response , key):
	response.delete_cookie(key)



# Generated by Django 3.0.3 on 2020-04-26 07:35

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('experiment', '0002_experimentgroup_project'),
    ]

    operations = [
        migrations.AlterField(
            model_name='project',
            name='name',
            field=models.CharField(max_length=50, unique=True),
        ),
    ]



'''
数据库对象代理，提供方便的（更方便的）访问数据库对象的方式
包括自动查找、自动保存等功能
'''
from .experi import Experiment , ExperimentGroup , Project
from .variable import Variable , VariableTrack , SingleValue
from .others import Figure

from .experi import SQL_Experiment , SQL_ExperimentGroup , SQL_Project
from .variable import SQL_Variable , SQL_VariableTrack , SQL_SingleValue
from .others import SQL_Figure


# Generated by Django 3.0.3 on 2020-04-29 14:18

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('experiment', '0009_auto_20200429_2147'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='variable',
            name='default_val',
        ),
    ]



from ...models import Variable
from ...utils.str_opt import seped_s2list , seped_list2s , seped_s2list_allow_empty

def save_editables(editable_id , editable_val):
	'''更新那些可修改变量的值。注意这里不会创建新SingleValue'''
	
	editable_id = [int(x) for x in seped_s2list(editable_id)]
	editable_val = seped_s2list_allow_empty(editable_val) #允许空白字符
	for k , v_id in enumerate(editable_id):
		if v_id < 0:
			continue
		varia = Variable.objects.get(id = v_id)

		track = varia.tracks.filter(name = "default")
		if len(track) <= 0: #没有default track，则跳过此变量
			continue
		track = track[0]

		val = track.values.latest("time_stamp")
		val.value = editable_val[k]
		val.save()



from ...models import Variable , VariableTrack , SingleValue
from ...utils.str_opt import seped_s2list , seped_list2s , seped_s2list_allow_empty
from .utils import rand_num

def append_ids(the_ids , heads , lines , extras , hidden_heads = []):
	'''为输出的表格添加id列'''
	assert len(the_ids) == len(lines)

	heads = ["id"] + heads
	lines = [ 

		[ ( rand_num() , the_ids[i]) ] # (第一个元素是随便加的)
		+ lines[i]  
		for i in range(len(lines))
	] 
	extras = [{"fixed": "left"}] + extras

	if "id" in hidden_heads:
		extras["hide"] =  True

	return heads , lines , extras

def get_head_reorder(heads , show_order):
	'''将head重排序为config中保存的顺序'''

	show_order = [x for x in show_order if x in heads] #按顺序挑出那些记录了的head

	child_pos = [] #哪些head位置是需要重排序的
	for i in range(len(heads)):	
		if heads[i] in show_order:
			child_pos.append(i)
	for i in range(len(child_pos)): #将这些位置替换成show_order
		heads[child_pos[i]] = show_order[i]

	return heads

def get_expe_reorder(expe_lis):
	'''重排序实验
		异常退出的实验一定排在后面
	'''

	expe_good = [exp for exp in expe_lis if exp.state != 3]
	expe_bad  = [exp for exp in expe_lis if exp.state == 3]

	return expe_good + expe_bad

def experiment_list_to_str_list(
		expe_lis , 
		hidden_heads 	= [] , 
		show_order 		= [] , 
		fixed_left 		= [] , 
		fixed_right 	= [] , 
	):
	'''输出前端需要的表格的各种信息
	'''
	heads 	= {}
	values 	= []
	lines 	= []
	extras 	= []
	editable = set()

	# 获得要显示的实验列表
	expe_lis = get_expe_reorder(expe_lis)

	# 获得表格的全部基本信息，包括表格头和表格值
	for exp in expe_lis:
		
		value_map = {}
		for varia in exp.variables.all():

			track = varia.tracks.filter(name = "default")
			if len(track) <= 0: #没有default track，则随机找一个track
				track = varia.tracks.all()
			if len(track) <= 0: #没有任何一个track，则跳过此变量
				continue
			track = track[0]

			if len( track.values.all() ) <= 0: #如果track没有变量，也跳过
				continue
				
			value_map[varia.name] = [ # value的每个元素是 (变量id，变量的值，变量的名)
				varia.id ,  
				track.values.latest("time_stamp").value , #找到这个track最新的一个变量
			]

			# 只要有一个元素可编辑，就整列可编辑
			if varia.editable:
				editable.add(varia.name)

		heads.update(value_map)
		values.append(value_map)

	# 获得 head 的排列顺序
	heads = get_head_reorder(list(heads) , show_order)

	# 获得每一行
	for i , exp in enumerate(expe_lis):
		this_line = []
		for h in heads:
			this_line.append( values[i].get(h , (rand_num() , "-")))
		lines.append(this_line)

	# 决定head的其余选项
	extras = [{} for _ in range(len(heads))]
	for i in range(len(heads)):
		if heads[i] in hidden_heads: # 前端点击隐藏的
			extras[i]["hide"] = True
		if heads[i] in fixed_left: # 固定左侧的
			extras[i]["fixed"] = "left"
		if heads[i] in fixed_right: # 固定右侧的
			extras[i]["fixed"] = "right"
		if heads[i] in editable:
		 	extras[i]["editable"] = True

	# 把id列添加进去
	ids = [exp.id for exp in expe_lis]
	heads , lines , extras = append_ids(ids , heads , lines , extras , hidden_heads)

	return ids , heads , lines , extras
		





from ..sql_proxy import Experiment 		as Core_Experiment
from ..sql_proxy import ExperimentGroup as Core_Group
from .others import Variable
from .paint import Painter
from ..quit import add_quit_process

this_experiment = None

def new_or_load_experiment(group_id = None, group_name = None, project_id = None, project_name = None , force_new = False):
	'''创建实验。如果已经创建过，就直接返回刚刚创建的那个实验
	
	参数：
		group_id：实验组的id
		group_name：实验组的名。和group_id只需给定其中一个
		project_id：项目id
		project_name：项目名。和project_id只需给出其中一个
		force_new：如果为True，则强制创建新实验。默认为False。
	'''
	global this_experiment
	if this_experiment is not None:
		if not force_new:
			return this_experiment

	new_exp = Experiment(group_id, group_name, project_id, project_name)
	this_experiment = new_exp

	return new_exp

def new_or_load_expr_from_cmd(arg_prx , args = None , force_new = False):
	'''从命令行参数直接创建实验。如果已经创建过，就直接返回刚刚创建的那个实验
	
	参数：
		arg_prx：超参数代理
		args：命令行参数列表。默认使用sys.args
		force_new：如果为True，则强制创建新实验。默认为False。
	'''
	global this_experiment
	if this_experiment is not None:
		if not force_new:
			return this_experiment

	C = arg_prx.assign_from_cmd(args)

	new_exp = Experiment(group_name = C.fitter_group, project_name = C.fitter_project)
	new_exp.use_argument_proxy(arg_prx , args)
	this_experiment = new_exp

	return new_exp


class Experiment:
	def __init__(self , group_id = None, group_name = None, project_id = None, project_name = None):
		'''创建实验。
		'''
		
		self._get_core(group_id , group_name , project_id , project_name)
		self.variables = {}
		self.figures = {}
		self.add_line = self.write_log
		self.id = int(self.core.id)

	def _get_core(self , group_id = None, group_name = None, project_id = None, project_name = None):

		if (group_id is None) and (group_name is None):
			group_name = "default"
		if (project_id is None) and (project_name is None):
			project_name = "default"
		self.core = Core_Experiment(group = Core_Group.find(group_id , group_name , project_id , project_name))

	def finish(self):
		'''结束实验。在实验正常结束时调用'''
		from django.utils import timezone
		self.core.state = 1
		self.core.end_time = timezone.now()

	def write_log(self , content = ""):
		'''写一行文字log'''
		content = str(content) + "\n"
		self.core.logs = self.core.logs + content

	def new_variable(self , name , type = str , default = "None" , merge_func = None , editable = False):
		'''新建一个变量
		参数：
			name：变量的名。请保证不重复。
			type：如何解释变量（所有变量都以字符串储存，这个函数将字符串解释成值）。
			default：变量的默认值。可以给字符串。
			merge_func：function。如何合并所有track。如果给定这个参数，则每次更新其他track时时会自动更新default。
			editable：是否可以在前端修改。
		'''
		self.variables[name] = Variable(self , name , type , default , merge_func , editable)

	def use_argument_proxy(self , arg_prox , args = None):
		'''使用一个参数代理
		'''
		C = arg_prox.assign_from_cmd(args)
		for arg in arg_prox.args:
			if arg.type.name == "_FITTER_SPLITTER":
				continue
			self.new_variable(arg.name , arg.type , str(C.__dict__[arg.name]) , editable = arg.editable)

	def new_figure(self , name):
		fig = Painter(name , self)
		self.figures[name] = Painter(name , self)
		return fig

	def __getitem__(self , name):
		return self.variables[name]



def make_exp_state_on_quit():
	if this_experiment is not None:
		if this_experiment.core.state == 0:
			print ("unexpected quit!!!")
			this_experiment.core.state = 3 #unexpected quit

add_quit_process(make_exp_state_on_quit)



from .base import Object 
from .utils import make_obj_list , none_or_id
import time
from ..importer import Figure as SQL_Figure

class Figure(Object):
	'''一个图片
	'''

	def __init__(self , name = None , experiment = None , from_obj = None):

		super().__init__(SQL_Figure , from_obj , name = name , expe_id = none_or_id(experiment))

		self.set_name_map(
			name 		= "name" , 
			html 		= "html" ,
			sql_expe 	= "expe" ,
			id 			= "id" , 
		)



__all__ = [
	"Figure" 			, "SQL_Figure" 		, 
]


import os

def read_config(config_path):
	if not os.path.exists(config_path):
		raise "file does not exists."

	nspace = {}
	with open(config_path , "r") as fil:
		config_content = fil.read()

	exec(config_content , nspace)
	argprox = nspace["get_arg_proxy"]()

	return argprox



from django.shortcuts import render
from django.http import HttpResponse , Http404 , HttpResponseRedirect
from ...models import Project , ExperimentGroup , Experiment
from ...models import Variable , VariableTrack , SingleValue
from ..base import get_path
from .utils import *

def variable(request , variable_id):

	variable = Variable.objects.get(id = variable_id)

	tracks_and_values = get_tracks(variable.tracks.all())

	context = {
		"variable": variable ,
		"tracks_and_values": tracks_and_values ,  
	}
	return render(request , get_path("variable/variable") , context)



def new_variable(request , experiment_id):

	if request.POST:
		name = request.POST.get("name")
		vari = Variable(name = name , expe_id = experiment_id)
		vari.save()
	return HttpResponseRedirect("/experiment/%s" % (str(experiment_id)))


def track(request , track_id):

	track = VariableTrack.objects.get(id = track_id)

	context = {
		"track": track , 
		"values": track.values.all() , 
	}
	return render(request , get_path("track") , context)



import os , sys
import os.path as P
import django


target_path = P.join( P.dirname(__file__) , "../fitterlog_server_module/")
flag = target_path in sys.path

if not flag:
	sys.path.append( target_path )

os.environ.setdefault('DJANGO_SETTINGS_MODULE', 'fitterlog_server.settings')
django.setup()

from experiment.models import *

if not flag:
	sys.path.remove( target_path )




# Generated by Django 3.0.3 on 2020-04-27 17:41

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('experiment', '0005_auto_20200428_0122'),
    ]

    operations = [
        migrations.AlterField(
            model_name='experiment',
            name='start_time',
            field=models.DateTimeField(auto_now=True),
        ),
        migrations.AlterField(
            model_name='experimentgroup',
            name='project',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='groups', to='experiment.Project'),
        ),
    ]



short_name_len = 50
path_len = 200
value_len = 100

special_postfix = "__FITTERLOG__"
split_token = "__FITTERLOG__SEP__"
config_file_key = "__FITTERLOG__CONFIG__"
cookie_key = "FITTERLOG_PERMISSION"

# 如果为True，会屏蔽登录机制。否则每次reload都要登录。
debugging = True


from django.shortcuts import render
from django.http import HttpResponse , Http404 , HttpResponseRedirect
from ...models import Project , ExperimentGroup
from ..base import get_path
from .utils import group_sort , get_num_exp
from ...utils.permission import check_permission
from ..displays import ask_login
from ...utils.str_opt import seped_s2list , seped_list2s
from .create import *
from .search import *

def project(request , project_id):
	if not check_permission(request):
		return ask_login(request)

	project = Project.objects.get(id = project_id)
	project.checkconfig()

	groups = group_sort(project.groups.all())
	inform = get_num_exp(groups) 

	context = {
		"project": project , 
		"groups" : groups, 
		"inform" : inform, 
		"config_files" : seped_s2list(project.config_files), 
	}
	return render(request , get_path("project/project") , context)


def new_project(request):
	if not check_permission(request):
		return ask_login(request)

	if request.POST:
		name = request.POST.get("name")
		path = request.POST.get("path")

		proj = Project(name = name , path = path)
		proj.save()
		grop = ExperimentGroup(name = "default", project_id = proj.id)
		grop.save()
	return HttpResponseRedirect("/")

def project_save_config(request , project_id):
	if not check_permission(request):
		return ask_login(request)

	project = Project.objects.get(id = project_id)
	project.checkconfig()

	if request.POST:
		for name in request.POST:
			content = request.POST[name]
			name = name.strip()

			if name == "project-intro":
				project.intro = content
			if name == "project-path":
				project.path = content
			if name == "config-files":
				project.config.config_files = content

			if name == "cmd-pref":
				project.config.cmd_pref = content
			if name == "cmd-comm":
				project.config.cmd_comm = content
			if name == "cmd-entr":
				project.config.cmd_entr = content
			if name == "cmd-suff":
				project.config.cmd_suff = content

			if name.startswith("group-intro-"):
				grp_id = int(name.split("group-intro-")[1])
				grp = ExperimentGroup.objects.get(id = grp_id)
				grp.intro = content
				grp.save()
		project.save()

	return HttpResponseRedirect("/project/%d" % project_id)


def smooth_lis(lis):
	lim = 500
	if len(lis) < lim:
		return lis
	p = len(lis) / lim #每p个采样一次

	lis = [lis[int(p*i+0.5)] for i in range(lim) if int(p*i+0.5) < len(lis)]

	return lis

def get_tracks(track_lis):

	t_vs = []

	for t in track_lis:
		vlis = []
		for v in t.values.all():
			if v.time_stamp < 0:
				continue
			val = v.value
			try:
				float(val)
			except ValueError: #not a float
				continue
			vlis.append( (v.time_stamp , val) )
		vlis = smooth_lis(vlis)
		t_vs.append( (t , vlis))

	return t_vs



# Generated by Django 3.0.3 on 2020-05-18 02:28

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('experiment', '0018_figure'),
    ]

    operations = [
        migrations.AlterField(
            model_name='figure',
            name='name',
            field=models.CharField(max_length=50),
        ),
    ]



# Generated by Django 3.0.3 on 2020-05-07 02:21

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('experiment', '0015_auto_20200507_0929'),
    ]

    operations = [
        migrations.AddField(
            model_name='groupconfig',
            name='hide_bad_exp',
            field=models.BooleanField(default=False),
        ),
    ]



import threading
from .cmd_start import run_a_experiment
import pynvml
import time

pynvml_failed = False
try:
	pynvml.nvmlInit()
except pynvml.nvml.NVMLError_LibraryNotFound:
	pynvml_failed = True

class ExperimentProxer_CPU(threading.Thread):
	def __init__(self , wait_time = 0):
		threading.Thread.__init__(self)
		self.setDaemon(True)

		self.tasks = []
		self.protect = True #设置这个变量来防止代理器还没添加任务就被意外删除。在添加任务完毕后关闭保护。
		self.wait_time = wait_time
		self.closed = False
	def add_task(self , path , config_name , values , 
			command = "python" , entry_file = "main.py" , prefix = "", suffix = ""):
		self.tasks.append( [path , config_name , values , command , entry_file , prefix, suffix] )

	def run(self):
		while not self.closed:
			time.sleep(self.wait_time + 1) # 等待多一秒，防止并发错误

			if len(self.tasks) <= 0:
				continue
			task = self.tasks.pop(0)

			run_a_experiment(*task)

	def close(self): #结束任务
		self.closed = True

class ExperimentProxer_Torch(threading.Thread):

	def __init__(self , gpus = [0] , max_process = 1 , wait_time = 0):
		threading.Thread.__init__(self)
		self.setDaemon(True)

		self.tasks = []
		self.gpus = gpus
		self.max_process = max_process
		self.protect = True
		self.wait_time = wait_time

		self.closed = False

		self._cnt = 0

	def good_gpu(self , k):
		if pynvml_failed:
			return self._cnt < self.max_process

			raise "PYNVML not useable"
		handle = pynvml.nvmlDeviceGetHandleByIndex(k)
		process = pynvml.nvmlDeviceGetComputeRunningProcesses(handle)

		return len(process) < self.max_process

	def add_task(self , path , config_name , values , 
			command = "python" , entry_file = "main.py" , prefix = "", suffix = ""):
		self.tasks.append( [path , config_name , values , command , entry_file , prefix, suffix] )

	def start_task_on_device(self , task , device):
		self._cnt += 1
		task[-2] = "CUDA_VISIBLE_DEVICES={0}".format(device) #前缀
		run_a_experiment(*task)

	def run(self):
		while not self.closed:
			time.sleep(self.wait_time + 1) # 防止并发错误

			if len(self.tasks) <= 0:
				continue

			task = self.tasks[0]

			flag = False # 是否成功运行了这个任务
			for k in self.gpus:
				if self.good_gpu(k): 
					self.start_task_on_device(task , k)
					flag = True
					break
			if flag:
				self.tasks.pop(0) #成功运行任务，删除此任务
	def close(self):
		self.closed = True

# expe_prox_cpu = ExperimentProxer_CPU()
# expe_prox_torch = ExperimentProxer_Torch()
# 
# expe_prox_cpu.start()
# expe_prox_torch.start()


def find_or_new(sql_class , **kwargs):
	'''根据 kwargs 规定的属性查找 sql 对象，如果找不到就新建一个。
	'''
	kwargs = {x : kwargs[x] for x in kwargs if kwargs[x] is not None } #去除None（不查找项）
	obj = sql_class.objects.filter(**kwargs)
	if len(obj) >= 1:
		return obj[0]
	return sql_class(**kwargs)

def make_obj_list(clas , sql_objs):
	'''根据给定的 sql 对象生成 fitterlog 对象
	'''
	return [clas(from_obj = obj) for obj in sql_objs]

def none_or_id(obj):
	return obj.id if obj is not None else obj


from .base import Object 
from .utils import make_obj_list , none_or_id
import time
from ..importer import Project as SQL_Project
from ..importer import Experiment as SQL_Experiment
from ..importer import ExperimentGroup as SQL_ExperimentGroup
from .variable import Variable

class Project(Object):
	'''一个项目
	'''

	def __init__(self , name = None , path = None , from_obj = None):

		super().__init__(SQL_Project , from_obj , name = name , path = path)

		self.set_name_map(
			name 		= "name" , 
			path 		= "path" ,
			sql_groups 	= "groups" ,
			id 			= "id" , 
		)

	def find(id = None , name = None):
		assert (id is not None) or (name is not None)
		if id is None:

			objs = SQL_Project.objects.filter(name = name)
			if len(objs) <= 0:
				proj = SQL_Project(name = name)
				proj.save()
			else:
				proj = objs[0]
			return proj

		return SQL_Project.objects.get(id = id)

class ExperimentGroup(Object):
	'''一组实验
	'''

	def __init__(self , name = None , project = None , from_obj = None):
		super().__init__(SQL_ExperimentGroup , from_obj , name = name , project_id = none_or_id(project))

		self.set_name_map(
			name 			= "name" , 
			id 				= "id" , 
			sql_experiments = "experiments" , 
			sql_project 	= "project" , 
		)

	def find(id = None , name = None , project_id = None , project_name = None):
		proj = Project.find(id = project_id , name = project_name)
		assert (id is not None) or (name is not None)

		if id is None:
			objs = SQL_ExperimentGroup.objects.filter(name = name , project = proj)
			if len(objs) <= 0:
				grop = SQL_ExperimentGroup(name = name , project = proj)
				grop.save()
			else:
				grop = objs[0]
			return grop
		return SQL_ExperimentGroup.objects.get(id = id , project = proj)

class Experiment(Object):
	'''一次实验
	'''

	def __init__(self , group = None , from_obj = None):

		super().__init__(SQL_Experiment , from_obj , force_new = True , group_id = none_or_id(group))

		self.set_name_map(
			logs 			= "logs" , 
			sql_variables 	= "variables" ,
			id 				= "id" , 
			state 			= "state" , 
			end_time 		= "end_time" , 
			start_time 		= "start_time" , 
		)



__all__ = [
	"Project" 			, "SQL_Project" 		, 
	"ExperimentGroup" 	, "SQL_ExperimentGroup" , 
	"Experiment" 		, "SQL_Experiment" 		, 
]


# Generated by Django 3.0.3 on 2020-07-28 12:47

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('experiment', '0020_project_config_files'),
    ]

    operations = [
        migrations.CreateModel(
            name='ProjectConfig',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('config_files', models.TextField(default='')),
                ('cmd_pref', models.CharField(default='', max_length=50)),
                ('cmd_comm', models.CharField(default='', max_length=50)),
                ('cmd_entr', models.CharField(default='', max_length=50)),
                ('cmd_suff', models.CharField(default='', max_length=50)),
            ],
        ),
        migrations.RemoveField(
            model_name='project',
            name='config_files',
        ),
        migrations.AddField(
            model_name='project',
            name='config',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, to='experiment.ProjectConfig'),
        ),
    ]



from django.shortcuts import render
from django.http import HttpResponse , Http404
from ...models import Experiment
from ..base import get_path

def experiment(request , experiment_id):

	experiment = Experiment.objects.get(id = experiment_id)

	context = {
		"experiment": experiment , 
		"variables": experiment.variables.all() , 
	}
	return render(request , get_path("experiment/experiment") , context)



def experiment_log(request , experiment_id):

	experiment = Experiment.objects.get(id = experiment_id)

	context = {
		"experiment": experiment , 
		"logs": experiment.logs , 
	}
	return render(request , get_path("experiment/logs") , context)

def experiment_figure(request , experiment_id):
	experiment = Experiment.objects.get(id = experiment_id)

	figs = [ [fig.name , fig.html] for fig in experiment.figures.all()]
	context = {
		"experiment": experiment , 
		"figs": figs ,
		"first_name" : figs[0][0] if len(figs) > 0 else "" ,  
	}
	return render(request , get_path("experiment/figures") , context)



# Generated by Django 3.0.3 on 2020-05-04 08:48

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('experiment', '0011_auto_20200501_0145'),
    ]

    operations = [
        migrations.AddField(
            model_name='experiment',
            name='end_time',
            field=models.DateTimeField(null=True),
        ),
        migrations.AddField(
            model_name='experiment',
            name='running',
            field=models.BooleanField(default=True),
        ),
        migrations.AlterField(
            model_name='experiment',
            name='start_time',
            field=models.DateTimeField(auto_now_add=True),
        ),
    ]



from YTools.universe.strlen import max_len
import random

def rand_num():
	return random.randint(-1e8 , 0)

def generate_len(heads , lines):
	'''为heads生成长度（根据字符串长度）用于前端'''
	if len(lines) == 0:
		return []
	lens = [max_len(s) for s in heads]
	lens = [max( lens[k] , max( [max_len(line[k][1]) for line in lines] )) for k in range(len((heads)))]
	lens = [ min(50 + x*10 , 300) for x in lens]

	return lens



from django.urls import path
from . import views
from functools import partial

urlpatterns = [
	path("" 												, views.index),
	path("login" 											, views.login),
	path("new_project" 										, views.new_project),

	path("project/<int:project_id>" 						, views.project),
	path("project/<int:project_id>/new_group" 				, views.new_group),
	path("project/<int:project_id>/save_config" 			, views.project_save_config),
	path("project/<int:project_id>/create_experiment" 		, partial(views.experiment_to_create , experiment_id = None) ),
	path("project/<int:project_id>/new_experiment" 			, views.new_experiment),
	path("project/<int:project_id>/hyper_search" 			, views.hyper_search),

	path("group/<int:group_id>" 							, views.group),
	path("group/<int:group_id>/save_config" 				, views.group_save_config),
	path("group/<int:group_id>/get_data" 					, views.get_data),

	path("experiment/<int:experiment_id>" 					, views.experiment),
	path("experiment/<int:experiment_id>/new_variable" 		, views.new_variable),
	path("experiment/<int:experiment_id>/logs" 				, views.experiment_log),
	path("experiment/<int:experiment_id>/figures" 			, views.experiment_figure),
	path("experiment/<int:experiment_id>/copy" 				, partial(views.experiment_to_create , project_id = None) ),

	path("variable/<str:variable_id>" 						, views.variable),

	path("track/<str:track_id>" 							, views.track),

]


from .displays import *
from .group_opt import *
from .variable_opt import *
from .project_opt import *
from .experiment_opt import *

def test_data(request):
	from django.http import HttpResponse , Http404 , HttpResponseRedirect

	return HttpResponse("""{
		"code":0,
		"msg":"",
		"count":100,
		"data":[
			{"asd" : "233" , "SS" : "tr"} , 
			{"asd" : "23" , "SS" : "4"} 
		]
	}""")


from .base import Object
from .utils import make_obj_list , none_or_id
import time
from ..importer import SingleValue as SQL_SingleValue
from ..importer import VariableTrack as SQL_VariableTrack
from ..importer import Variable as SQL_Variable

class Variable(Object):

	def __init__(self , name = None , experiment = None , from_obj = None):
		super().__init__(SQL_Variable , from_obj , name = name , expe_id = none_or_id(experiment))
		
		self.set_name_map(
			name 		= "name" , 
			experiment 	= "experiment" ,
			sql_tracks 	= "tracks" ,
			id 			= "id" , 
			editable    = "editable" , 
		)


class VariableTrack(Object):
	'''某个变量的一条时间线
	'''

	def __init__(self , name = None , variable = None , from_obj = None):
		super().__init__(SQL_VariableTrack , from_obj , name = name , variable_id = none_or_id(variable))
		self.set_name_map(
			name 		= "name" , 
			variable 	= "variable" ,
			sql_values 	= "values" ,
			id 			= "id" , 
		)

class SingleValue(Object):
	'''某个变量的某条时间线上的一个值
	'''

	def __init__(self , value = None, timestamp = None , track = None , from_obj = None):

		super().__init__(SQL_SingleValue , from_obj , value = value, time_stamp = timestamp , track_id = none_or_id(track))
		self.set_name_map(
			value 		= "value" , 
			timestamp 	= "time_stamp" ,
			id 			= "id" , 
		)


__all__ = [
	"SingleValue" , 
	"VariableTrack" , 
	"Variable" , 
]
__all__ = [
]


from .utils import *
import os
from subprocess import Popen , DEVNULL
from contextlib import redirect_stdout

def run_a_experiment(path , config_name , values , 
		command = "python" , entry_file = "main.py" , prefix = "", suffix = ""):
	from fitterlog.arg_proxy.types import Bool
	'''
		config_name: str。 对应的config文件的名。用于获得各个参数的类型。
		values: dict。 名-值字典。
	'''

	argprox = read_config(os.path.join(path , config_name))

	cmds = []
	for name , value in values.items():
		name = argprox.process_name(name)

		the_arg = argprox.name2arg.get(name)
		if the_arg is None:
			continue
		if the_arg.type is Bool:
			if not Bool(value):
				cmd = ""
			else:
				cmd = "--{name}".format(name = name)
		else:
			# "" is always acceptable while '' is only for linux.
			cmd = "--{name}=\"{value}\"".format(name = name , value = value) 

		cmds.append(cmd)

	cmd = "{prefix} {command} {entry_file} {cmds} {suffix}".format(
		prefix = prefix , command = command , entry_file = entry_file , cmds = " ".join(cmds) , suffix = suffix , 
	)

	# && is always acceptable while & is only for windows
	full_cmd = "cd {path} && {cmd}".format(path = path , cmd = cmd) 

	return Popen(full_cmd , shell = True , stdout = DEVNULL , stderr = DEVNULL)



# Generated by Django 3.0.3 on 2020-05-04 13:13

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('experiment', '0013_experimentgroup_start_time'),
    ]

    operations = [
        migrations.AddField(
            model_name='groupconfig',
            name='show_order',
            field=models.TextField(default=''),
        ),
    ]



from ..interface.others import Variable
import argparse
from .types import *
from YTools.universe import beautiful_str

class Argument:
	def __init__(self , name , type , default , editable = False):
		self.name = name
		self.type = type 
		self.default = default
		self.editable = editable

class ArgProxy:
	def __init__(self , name = None):
		self.name = name
		self.args = []
		self.name2arg = {}

		self.add_argument("fitter_project" , String , default = "default")
		self.add_argument("fitter_group" , String , default = "default")
		self.add_splitter()

	def process_name(self , name):
		return name.strip().replace(" " , "_")

	def add_splitter(self , _ = ""):
		self.args.append( Argument("" , FITTER_SPLITTER , None) )

	def add_argument(self , name , type = String , default = None , editable = False):
		name = self.process_name(name)
		arg = Argument(name , type , default , editable = editable)
		self.args.append(arg)
		self.name2arg[name] = arg

	def add_store_true(self , name , editable = False):
		name = self.process_name(name)
		arg = Argument(name , Bool , "False" , editable = editable)
		self.args.append(arg)
		self.name2arg[name] = arg

	def assign_from_cmd(self , args = None):
		C = argparse.ArgumentParser()

		for arg in self.args:
			if arg.type.name == "_FITTER_SPLITTER":
				continue 

			if arg.type == Bool:
				C.add_argument("--%s" % arg.name , action = "store_true" , default = False)
			else:
				C.add_argument("--%s" % arg.name , type = arg.type , default = arg.default)

		C = C.parse_args(args)

		return C

	def __str__(self):
		return beautiful_str(["name" , "default" , "type"] , [
			[x.name , x.default , x.type] for x in self.args
		])




from django.shortcuts import render
from django.http import HttpResponse , Http404 , HttpResponseRedirect
from ..models import Project
from .base import get_path
from ..utils.permission import check_permission as check_per , give_permission as give_per
from ..utils.permission import check_passwords
from urllib.parse import urlencode

def index(request):
	context = {
		"projects": Project.objects.all() , 
	}
	return render(request , get_path("index") , context)

def ask_login(request , path = None , info = ""):

	if path is None:
		path = request.get_full_path()
	return render(request , get_path("ask_login") , {
		"path": path , 
		"info": info , 
	})

def login(request):
	if request.POST:
		password = request.POST.get("password")
		path = request.POST.get("path")

		if password in check_passwords():
			return give_per(request , HttpResponseRedirect(path))
		return ask_login(request , path = path ,  info = "不对啊啊啊啊啊啊")

	raise Http404


from django.shortcuts import render
from django.http import HttpResponse , Http404 , HttpResponseRedirect
from ...models import Project , ExperimentGroup , Experiment
from ...models import Variable , VariableTrack , SingleValue
from ..base import get_path
from ...utils.str_opt import seped_s2list , seped_list2s
from .editable import save_editables
from .ask_database import experiment_list_to_str_list
from .fake_frontface import cols_from_data
import json
from django.db.models.functions import Floor
from django.db.models import F
import django

def find_range_explist(group , page , limit , hide_ids):
	'''获取某一个特定范围的实验列表，而不是全部'''

	l = (page-1) * limit # 注意page是从1开始的
	r = page * limit

	# 先排除删除的id，再排序，再取区间
	# 第一个排序把先运行的实验放到后面
	# 第二个排序把异常终止的实验（state=3）放到后面
	return group.experiments.exclude(id__in = hide_ids).order_by( Floor(F("state") / 3) , "-start_time")[l : r]


def get_data(request , group_id):

	''' 把表格数据传过去
	'''
	group = ExperimentGroup.objects.get(id = group_id)
	group.checkconfig()

	get_parameter 	= json.loads( list(request.GET)[0] ) # layui的傻逼实现
	page 			= get_parameter["page"] #页号，从1开始
	limit 			= get_parameter["limit"] #每页数量
	
	# 获得保存的各种设置
	hide_heads 	= seped_s2list(group.config.hidden_heads) 	# 隐藏的列
	hide_ids 	= [int(x) for x in seped_s2list(group.config.hidden_ids) if x.isdigit()] # 隐藏的行
	show_order 	= seped_s2list(group.config.show_order)  	# 列顺序
	fixed_left 	= seped_s2list(group.config.fixed_left)  	# 左侧固定的列
	fixed_right = seped_s2list(group.config.fixed_right) 	# 右侧固定的列

	# 生成行和列信息
	expe_list = find_range_explist(group , page , limit , hide_ids) #这一步已经去除了所有隐藏的行

	ids , heads , lines , extras = experiment_list_to_str_list( 
		expe_list , hide_heads , show_order , fixed_left , fixed_right
	)

	# 生成要传回的data
	datas = []
	for i , line in enumerate(lines):
		exp_id 	= ids[i]
		state 	= Experiment.objects.get(id = exp_id).state

		val_map = {}
		vid_map = {}

		for i , (v_id , val) in enumerate( line ):

			val_map [heads[i]] = val
			vid_map [heads[i]] = v_id

		datas.append({
			"exp_id" 	: exp_id , 
			"state"  	: state , 
			"val" 		: val_map , 
			"vid" 		: vid_map , 
		})
	
	# 生成要返回的列名
	cols = cols_from_data(heads , lines , extras)

	# 生成额外信息
	extra_elements = {
		"editable": { heads[i]: (extras[i].get("editable") is not None) for i in range(len(heads))}, 
		"page" 	  : page ,
		"limit"	  : limit ,
		"tot_num" : group.experiments.exclude(id__in = hide_ids).count() , 
	}

	# 生成返回的json字符串
	resp = json.dumps({
		"data": datas , 
		"cols": cols , 
		"extr": extra_elements , 
	})

	return HttpResponse(resp)




from django.shortcuts import render
from django.http import HttpResponse , Http404 , HttpResponseRedirect
from ...models import Experiment , Project
from ..base import get_path
import os
from fitterlog_cmd.cmd_start import run_a_experiment
from ...utils.constants import special_postfix , config_file_key
from ...utils.permission import check_permission
from ..displays import ask_login

def type2str(type):
	if hasattr(type , "name"):
		return type.name
	return "unknown"

def get_val(exp , name , default = ""):
	var = exp.variables.filter(name = name)
	if len(var) <= 0:
		return str(default)
	var = var[0]

	trk = var.tracks.filter(name = "default")
	if len(trk) < 0:
		return str(default)
	trk = trk[0]

	return trk.values.latest("time_stamp").value

def experiment_to_create(request , project_id = None, experiment_id = None):
	'''新建实验的界面'''

	# 要求权限
	if not check_permission(request):
		return ask_login(request)

	# 至少要有一个参数
	if project_id is None and experiment_id is None:
		raise Http404

	# 获取对象
	if experiment_id is not None:
		experiment = Experiment.objects.get(id = experiment_id)
		project = experiment.group.project
	else:
		project = Project.objects.get(id = project_id)
	project.checkconfig()

	# 获取config文件名
	if request.POST:
		config_name = request.POST.get("chosen-config")
	else:
		raise Http404
	
	# 获取config文件
	target_file = os.path.join(project.path , config_name)
	if not os.path.exists(target_file):
		raise Http404

	# 运行config文件，获取argprox对象
	nspace = {}
	with open(target_file , "r") as fil:
		config_content = fil.read()

	exec(config_content , nspace)
	argprox = nspace["get_arg_proxy"]()

	# 决定各种参数
	if experiment_id is not None: # 复制实验
		args = [ [type2str(x.type) , x.name , get_val(experiment , x.name , x.default).strip()] 
					for x in argprox.args]
		last_page_path = "/group/{0}".format(experiment.group.id)
	else: #使用默认值
		args = [ [type2str(x.type) , x.name , x.default] for x in argprox.args]
		last_page_path = "/project/{0}".format(project.id)

	# 返回的上下文
	context = {
		"args" : args , 
		"project" : project , 
		"config_file": target_file , 
		"config_name": config_name , 
		"last_page_path": last_page_path,
	}

	return render(request , get_path("project/create/experiment_create") , context)


def new_experiment(request , project_id):
	'''根据获得的各种信息来在命令行开始一个实验（运行代码）。'''	

	if not check_permission(request):
		return ask_login(request)

	project = Project.objects.get(id = project_id)


	values = {}
	if request.POST:
		for name in request.POST:
			if name == "csrfmiddlewaretoken":
				continue
			if name.startswith(special_postfix):
				continue
			values[name] = str(request.POST[name])
		config_name = request.POST[config_file_key]
		last_page_path = request.POST["__last_page_path"]

		prefix 	= request.POST["__prefix"]
		command = request.POST["__command"]
		file 	= request.POST["__file"]
		suffix 	= request.POST["__suffix"]

	run_a_experiment(
		path 		= project.path , 
		config_name = config_name , 
		values 		= values, 
		command 	= command , 
		entry_file 	= file , 
		prefix 		= prefix , 
		suffix 		= suffix , 
	)
	
	return HttpResponseRedirect(last_page_path)


#!/usr/bin/env python
"""Django's command-line utility for administrative tasks."""
import os
import sys


def main():
	os.environ.setdefault('DJANGO_SETTINGS_MODULE', 'fitterlog_server.settings')
	try:
		from django.core.management import execute_from_command_line
	except ImportError as exc:
		raise ImportError(
			"Couldn't import Django. Are you sure it's installed and "
			"available on your PYTHONPATH environment variable? Did you "
			"forget to activate a virtual environment?"
		) from exc
	execute_from_command_line(sys.argv)


if __name__ == '__main__':
	main()



from django.db import models
from ..utils.constants import short_name_len , path_len
from ..utils.str_opt import seped_list2s , seped_s2list
from .configs import GroupConfig , ProjectConfig

class Project(models.Model):
	'''一个项目
	'''	
	name = models.CharField(max_length = short_name_len , unique = True)
	path = models.CharField(max_length = path_len)
	intro = models.TextField(default = "")
	config = models.OneToOneField(ProjectConfig , on_delete = models.SET_NULL , null = True)

	def __str__(self):
		return self.name

	def checkconfig(self):
		if self.config is None:
			self.config = ProjectConfig()
			self.config.save()
			self.save()

	def save(self):
		self.checkconfig()
		self.config.save()
		return super().save()

	@property
	def config_files(self):
		self.checkconfig()
		return self.config.config_files

	@config_files.setter
	def config_files(self , v):
		self.checkconfig()
		self.config.config_files = v
	

class Experiment(models.Model):
	'''一次实验
	'''
	logs = models.TextField(default = "")
	intro = models.TextField(default = "")
	group = models.ForeignKey("ExperimentGroup" , on_delete = models.CASCADE , related_name = "experiments")

	# 0: running , 1: finished , 2: almost finished , 3: unexpected quit
	state = models.IntegerField(default = 0)

	start_time = models.DateTimeField(auto_now_add = True) # 创建时间
	end_time   = models.DateTimeField(null = True)

	def __str__(self):
		return self.name

class ExperimentGroup(models.Model):
	name = models.CharField(max_length = short_name_len)
	intro = models.TextField(default = "")
	project = models.ForeignKey(Project , on_delete = models.CASCADE , related_name = "groups")
	config = models.OneToOneField(GroupConfig , on_delete = models.SET_NULL , null = True)

	start_time = models.DateTimeField(auto_now_add = True) # 创建时间

	def __str__(self):
		return self.name

	def checkconfig(self):
		if self.config is None:
			self.config = GroupConfig()
			self.config.save()

	def save(self):
		self.checkconfig()
		self.config.save()
		return super().save()

	def add_hide_cols(self , s):
		self.checkconfig()

		hide_cols = seped_s2list(s)
		hide_cols = list(set(hide_cols))
		self.config.hidden_heads = seped_list2s(hide_cols)

	def add_hide_ids(self , s):
		self.checkconfig()

		hide_ids = seped_s2list(self.config.hidden_ids)
		hide_ids += seped_s2list(s)
		hide_ids = list(set(hide_ids))
		self.config.hidden_ids = seped_list2s(hide_ids)


__all__ = [
	"Project" , 
	"Experiment" , 
	"ExperimentGroup" , 
]


import os
import sys
import argparse

def migrate():
	os.chdir(os.path.dirname(__file__))
	sys.path.append(os.path.abspath("."))

	if os.system("python manage.py migrate") > 0:
		os.system("python3 manage.py migrate")

def run():
	os.chdir(os.path.dirname(__file__))
	sys.path.append(os.path.abspath("."))

	C = argparse.ArgumentParser()
	C.add_argument("--port" , type = int , default = 8000)
	C = C.parse_args()

	if os.system("python manage.py migrate") > 0:
		os.system("python3 manage.py migrate")

	if os.system("python manage.py runserver 0.0.0.0:%d" % (C.port)) > 0:
		os.system("python3 manage.py runserver 0.0.0.0:%d" % (C.port))


# Generated by Django 3.0.3 on 2020-05-18 02:23

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('experiment', '0017_variable_editable'),
    ]

    operations = [
        migrations.CreateModel(
            name='Figure',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=50, unique=True)),
                ('html', models.TextField(default='')),
                ('expe', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='figures', to='experiment.Experiment')),
            ],
        ),
    ]



from django.shortcuts import render
from django.http import HttpResponse , Http404 , HttpResponseRedirect
from ...models import Project , ExperimentGroup , Experiment
from ...models import Variable , VariableTrack , SingleValue
from ..base import get_path
from ...utils.str_opt import seped_s2list , seped_list2s
from .editable import save_editables
from .get_data import *

def group(request , group_id):

	group = ExperimentGroup.objects.get(id = group_id)
	
	context = {
		"group": group , 
		"config_files": seped_s2list(group.project.config_files) , 
	}
	return render(request , get_path("group/group") , context)

def new_group(request , project_id):

	if request.POST:
		name = request.POST.get("name")

		grop = ExperimentGroup(name = name , project_id = project_id)
		grop.save()
	return HttpResponseRedirect("/project/%s" % str(project_id))


def group_save_config(request , group_id):
	group = ExperimentGroup.objects.get(id = group_id)

	if request.POST:
		hide_columns = request.POST.get("hide_columns")
		hide_ids 	 = request.POST.get("hide_ids")
		intro 		 = request.POST.get("intro")
		show_order 	 = request.POST.get("show_order")
		hide_bad_exp = request.POST.get("hide_bad_exp")
		editable_id  = request.POST.get("editable_id") 
		editable_val = request.POST.get("editable_val") 
		fixed_left	 = request.POST.get("fixed_left") 
		fixed_right	 = request.POST.get("fixed_right") 

		save_editables(editable_id , editable_val)


		hide_bad_exp = True if hide_bad_exp == "true" else False

		group.checkconfig()

		group.config.hide_bad_exp = hide_bad_exp
		group.add_hide_cols(hide_columns)
		group.add_hide_ids(hide_ids)
		group.intro = intro
		group.config.show_order = show_order
		group.config.fixed_left = fixed_left
		group.config.fixed_right = fixed_right
		group.save()

	return HttpResponseRedirect("/group/%s" % (str(group_id)))



# Generated by Django 3.0.3 on 2020-04-30 17:45

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('experiment', '0010_remove_variable_default_val'),
    ]

    operations = [
        migrations.CreateModel(
            name='GroupConfig',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('hidden_heads', models.TextField(default='')),
                ('hidden_ids', models.TextField(default='')),
            ],
        ),
        migrations.AddField(
            model_name='experimentgroup',
            name='config',
            field=models.OneToOneField(null=True, on_delete=django.db.models.deletion.SET_NULL, to='experiment.GroupConfig'),
        ),
    ]



# Generated by Django 3.0.3 on 2020-05-08 16:20

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('experiment', '0016_groupconfig_hide_bad_exp'),
    ]

    operations = [
        migrations.AddField(
            model_name='variable',
            name='editable',
            field=models.BooleanField(default=False),
        ),
    ]



import pandas as pd
import matplotlib.pyplot as plt
from io import BytesIO
import base64
from ..sql_proxy import Figure as Core_Figure

class Painter:
	def __init__(self , name , expe):
		self.name = name
		self.expe = expe
		self.fig = Core_Figure(name = name , experiment = expe.core)

	def __enter__(self):
		plt.figure()
		plt.title(self.name)
		return self

	def __exit__(self , *pargs , **kwargs):
		self.fig.html = self.save_html()
		pass

	def save_html(self):
		'''将当前的plt内容保存成html img对象
		'''
		buf = BytesIO()
		plt.savefig(buf)  
		plot_data = buf.getvalue()

		img_src = base64.b64encode(plot_data)

		img_src = img_src.decode()
		img_src = "data:image/png;base64," + img_src
		ele = "<img src=\"%s\">" % img_src

		return ele



'''
	伪前端，因为js太麻烦了，所以我直接在python里处理好发过去。
'''
from .utils import generate_len

def cols_from_data(heads , lines , extras):

	lens = generate_len(heads , lines)
	cols = []

	for i , head in enumerate(heads):
		extra = extras[i]
		width = lens[i]

		col = {
			"field"		: "inner-{head}".format(head = head), 
			"title"		: head, 
			"width" 	: width , 
			"minWidth"	: width // 2 , 
			"sort" 		: True ,
			"align" 	:'center', 
		}
		if extra.get("hide" , False):
			col["hide"] = True
		if extra.get("fixed") is not None:
			col["fixed"] = extra["fixed"]

		cols.append(col)

	return cols



def group_sort(group_list):
	'''按创建时间降序排列，将default放在最前面'''
	group_list = list(group_list)
	group_list.sort(key = lambda x: x.start_time , reverse = True)

	didx = None
	for i in range(len(group_list)):
		if group_list[i].name == "default":
			didx = i
	if didx is not None:
		group_list = group_list[didx:didx+1] + group_list[:didx] + group_list[didx+1:]

	return group_list

def get_num_exp(group_list):

	ret = []
	for g in group_list:
		tot_num = 0
		state_num = [0] * 4

		for e in g.experiments.all():
			tot_num += 1
			state_num[e.state] += 1
		ret.append( [tot_num , state_num , g] )

	return ret


from django.db import models
from ..utils.constants import short_name_len , value_len
from .experi import Experiment

class SingleValue(models.Model):
	'''某个变量的某条时间线上的一个值
	'''
	value = models.CharField(max_length = value_len)
	time_stamp = models.IntegerField()
	track = models.ForeignKey("VariableTrack" , on_delete = models.CASCADE , related_name = "values")

class VariableTrack(models.Model):
	'''某个变量的一条时间线
	'''
	name = models.CharField(max_length = short_name_len)
	variable = models.ForeignKey("Variable" , on_delete = models.CASCADE , related_name = "tracks")

	def __str__(self):
		return self.name

class Variable(models.Model):
	name = models.CharField(max_length = short_name_len)
	expe = models.ForeignKey(Experiment , on_delete = models.CASCADE , related_name = "variables")
	
	#注意，editable虽然记在Variable中，但是是表头的属性。一列只要有单元格可编辑这一列就都是可编辑的。
	editable = models.BooleanField(default = False) 

	def __str__(self):
		return self.name

__all__ = [
	"SingleValue" , 
	"VariableTrack" , 
	"Variable" , 
]


# Generated by Django 3.0.3 on 2020-04-27 17:22

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('experiment', '0004_experiment_start_time'),
    ]

    operations = [
        migrations.AlterField(
            model_name='experimentgroup',
            name='project',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='groups', to='experiment.Project', unique=True),
        ),
    ]



from .constants import split_token


def seped_s2list_allow_empty(s , sep = split_token):
	if s == None:
		return []

	return s.strip().split(sep)


def seped_s2list(s , sep = split_token):
	if s == None:
		return []
	return list(filter(lambda x:x , s.strip().split(sep)))

def seped_list2s(l , sep = split_token):
	return sep.join(l)


# Generated by Django 3.0.3 on 2020-04-25 18:40

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Experiment',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('logs', models.TextField(default='')),
            ],
        ),
        migrations.CreateModel(
            name='ExperimentGroup',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=50)),
            ],
        ),
        migrations.CreateModel(
            name='Project',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=50)),
                ('path', models.CharField(max_length=200)),
            ],
        ),
        migrations.CreateModel(
            name='Variable',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=50)),
                ('expe', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='variables', to='experiment.Experiment')),
            ],
        ),
        migrations.CreateModel(
            name='VariableTrack',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=50)),
                ('variable', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='tracks', to='experiment.Variable')),
            ],
        ),
        migrations.CreateModel(
            name='SingleValue',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('value', models.CharField(max_length=50)),
                ('time_stamp', models.IntegerField()),
                ('track', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='values', to='experiment.VariableTrack')),
            ],
        ),
        migrations.AddField(
            model_name='experiment',
            name='group',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='experiments', to='experiment.ExperimentGroup'),
        ),
    ]



from django.db import models
from ..utils.constants import short_name_len , path_len

class GroupConfig(models.Model):
	hidden_heads = models.TextField(default = "")
	hidden_ids 	 = models.TextField(default = "")
	show_order 	 = models.TextField(default = "")
	hide_bad_exp = models.BooleanField(default = False)
	fixed_left 	 = models.TextField(default = "" , null = True)
	fixed_right  = models.TextField(default = "" , null = True)

class ProjectConfig(models.Model):
	config_files = models.TextField(default = "")
	cmd_pref = models.CharField(default = "" , max_length = short_name_len)
	cmd_comm = models.CharField(default = "" , max_length = short_name_len)
	cmd_entr = models.CharField(default = "" , max_length = short_name_len)
	cmd_suff = models.CharField(default = "" , max_length = short_name_len)

__all__ = [
	"GroupConfig" , 
	"ProjectConfig" , 
]


from setuptools import setup, find_packages
from setuptools.command.install import install
import os

with open("README.md" , encoding = "utf-8") as f:
	readme = f.read()

with open("LICENSE" , encoding = "utf-8") as f:
	license = f.read()

with open("requirements.txt", encoding = "utf-8") as f:
	reqs = f.read()

pkgs = [p for p in find_packages() if p.startswith("fitterlog")]
print(pkgs)

setup(
	name 								= "fitterlog",
	version 							= "0.3.5",
	url 								= "http://github.com/FFTYYY/fitterlog",
	description 						= "",
	license 							= "MIT",
	long_description 					= readme,
	long_description_content_type 		= "text/markdown",
	author 								= "Yang Yongyi",
	author_email 						= "yongyyang17@fudan.edu.cn",
	python_requires 					= ">=3.6",
	packages 							= pkgs,
	install_requires				 	= reqs.strip().split("\n"),
	entry_points			= {"console_scripts": [
		"fitterlog-start-server=fitterlog_server_module.start_server:run" ,
		"fitterlog-init=fitterlog_server_module.start_server:migrate" ,
	]},

	include_package_data = True ,
)



# Generated by Django 3.0.3 on 2020-07-28 12:51

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('experiment', '0021_auto_20200728_2047'),
    ]

    operations = [
        migrations.AlterField(
            model_name='project',
            name='config',
            field=models.OneToOneField(null=True, on_delete=django.db.models.deletion.SET_NULL, to='experiment.ProjectConfig'),
        ),
    ]



from ..sql_proxy import Variable 		as Core_Variable
from ..sql_proxy import VariableTrack 	as Core_Track
from ..sql_proxy import SingleValue 	as Core_SingleValue

class SingleValue:
	def __init__(self , track , timestamp , value):
		self.timestamp = timestamp
		self.value = str(value)
		self.track = track

		self.core = Core_SingleValue(str(value) , timestamp , track = track.core)

	def __str__(self):
		return self.value

class Track:

	#public
	def __init__(self , variable , name , type , default_value):
		self.variable = variable
		self.name = name
		self.type = type
		self.default_value = default_value

		self.core = Core_Track(name , variable = variable.core)
		
		self.max_time_stamp = -1
		self.time_stamps = []
		self.single_values = {}

		self.single_values[-1] = SingleValue(self , -1 , default_value)
		self.now_value = self.single_values[-1]

	def __getitem__(self , time_stamp):
		return self.single_values[time_stamp]

	def update(self , value , time_stamp = None):
		'''更新track。
		time_stamp如果留空，则默认为上一次的值加一。初始默认为0。
		'''
		if time_stamp is None:
			time_stamp = self.max_time_stamp + 1

		if time_stamp <= self.max_time_stamp: # bad timestamp
			return 

		self.max_time_stamp = time_stamp

		self.time_stamps.append(time_stamp)
		self.single_values[time_stamp] = SingleValue(self , time_stamp , str(value))
		self.now_value = self.single_values[time_stamp]

		# 更新变量的值
		if self.name != "default":
			self.variable.merge()

	@property
	def value(self):
		'''返回最近的value的值（解释）
		返回值类型：self.type
		'''
		return self.type(self.recent_value)

	@property
	def recent_value(self):
		'''返回最近的value（不解释）
		返回值类型：str
		'''
		if len(self.time_stamps) > 0:
			return self.kth_value(-1).value #返回最近的值
		return self.default_value

	def kth_value(self , k):
		'''离散化后的时间戳。返回第k个value解释前的值。
		'''
		return self.single_values[self.time_stamps[k]]
	
	def __str__(self):
		return str(self.value)


# variable 实际上是 track 的公有继承。
class Variable:
	def __init__(self , experiment , name , type , default_value , merge_func , editable):
		self.experiment = experiment
		self.name = name
		self.type = type
		self.default_value = str(default_value)
		self.merge_func = merge_func
		self.editable = editable

		self.core = Core_Variable(name , experiment = experiment.core)
		self.core.editable = editable
		self.tracks = {}

		self.new_track("default")

		self.id = int(self.core.id)

	def new_track(self , name , default_value = None):
		if default_value is None:
			default_value = self.default_value
		self.tracks[name] = Track(self , name , self.type , str(default_value))

	def merge(self): #合并所有track的值，生成default的值
		if self.merge_func is None:
			return
		v_list = []
		for t in self.tracks:
			v_list.append([self.tracks[t].max_time_stamp , self.tracks[t].value])

		new_val = self.merge_func(*v_list)
		if new_val is not None:
			self.update(new_val)

	def update(self , value , time_stamp = None , track = "default"):
		self[track].update(value , time_stamp)

	@property
	def value(self):
		'''返回最近的value的值
		返回值类型：self.type
		'''
		return self["default"].value

	def __getitem__(self , name):
		if name not in self.tracks:
			if isinstance(name , int): #time stamp
				return self.tracks["default"][name]
			else: #new track name
				self.new_track(name)
		return self.tracks[name]

	def __str__(self):
		return str(self["default"])



from django.shortcuts import render
from django.http import HttpResponse , Http404 , HttpResponseRedirect
from ...models import Project
from ..base import get_path
import os
from fitterlog_cmd.cmd_start import run_a_experiment
from fitterlog_cmd.experiment_prox import ExperimentProxer_CPU , ExperimentProxer_Torch
from ...utils.permission import check_permission
from ..displays import ask_login
import copy

def dfs(dic , namelist , k , now_arg):
	if k >= len(namelist):
		return [ now_arg ]

	now_name = namelist[k]

	args = []
	for y in dic[now_name]:
		new_arg = copy.copy(now_arg)
		new_arg[now_name] = y
		args += dfs(dic , namelist , k + 1 , new_arg)
	return args

exp_proxs = []

def check_proxs(): #弹出那些已经没有任务的prox
	to_rem = []
	for i in range(len(exp_proxs)):
		if len(exp_proxs[i].tasks) == 0 and not exp_proxs[i].protect:
			exp_proxs[i].close() #关闭线程
			to_rem.append(i) 	 #移除对象
	to_rem.reverse()
	for x in to_rem:
		exp_proxs.pop(x)

def hyper_search(request , project_id):

	global exp_proxs

	# 进行一次检查，弹出那些以前创建的，已经跑完了的子线程，防止内存无限增长
	check_proxs()

	# 要求权限
	if not check_permission(request):
		return ask_login(request)

	# 获取对象
	project = Project.objects.get(id = project_id)

	# 获取搜索空间文件名
	if request.POST:
		search_name = request.POST.get("search-space")
	else:
		raise Http404
	
	# 获取空间文件
	search_file = os.path.join(project.path , search_name)
	if not os.path.exists(search_file):
		raise Http404

	# 运行空间文件，获取搜索空间
	nspace = {}
	with open(search_file , "r") as fil:
		search_content = fil.read()

	exec(search_content , nspace)

	d = nspace["get_search_space"]()
	comm 		= d["command"]
	main_file	= d["entry_file"]
	cfg_file 	= d["config_file"]
	space 		= d["space"]
	is_torch 	= d["is_torch"]
	wait_time 	= d["wait_time"]
	if is_torch:
		gpus 			= d["gpus"]
		max_proc_num 	= d["max_proc_num"]

	#遍历搜索空间，获取所有参数列表
	args = dfs(space , list(space) , 0 , {})

	if is_torch:
		prox = ExperimentProxer_Torch(gpus , max_proc_num , wait_time = wait_time)
	else:
		prox = ExperimentProxer_CPU(wait_time = wait_time)

	for arg in args:
		prox.add_task(
			path 		= project.path , 
			config_name = cfg_file , 
			values 		= arg, 
			command 	= comm , 
			entry_file 	= main_file , 
			prefix 		= "" , 
			suffix 		= "" , 
		)
	prox.protect = False
	prox.start()
	exp_proxs.append(prox)

	return HttpResponseRedirect("/project/%s" % (str(project_id)))






# Generated by Django 3.0.3 on 2020-05-07 01:29

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('experiment', '0014_groupconfig_show_order'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='experiment',
            name='running',
        ),
        migrations.AddField(
            model_name='experiment',
            name='state',
            field=models.IntegerField(default=0),
        ),
    ]



# Generated by Django 3.0.3 on 2020-04-26 07:34

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('experiment', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='experimentgroup',
            name='project',
            field=models.ForeignKey(default=None, on_delete=django.db.models.deletion.CASCADE, related_name='groups', to='experiment.Project'),
            preserve_default=False,
        ),
    ]



from django.db import models
from ..utils.constants import short_name_len
from .experi import Experiment

class Figure(models.Model):
	name  = models.CharField(max_length = short_name_len)
	html = models.TextField(default = "")
	expe  = models.ForeignKey(Experiment , on_delete = models.CASCADE , related_name = "figures")

	def __str__(self):
		return "<img:" + self.name + ">"

__all__ = [
	"Figure" , 
]


# Generated by Django 3.0.3 on 2020-04-29 04:58

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('experiment', '0006_auto_20200428_0141'),
    ]

    operations = [
        migrations.AddField(
            model_name='experiment',
            name='intro',
            field=models.TextField(default=''),
        ),
        migrations.AddField(
            model_name='experimentgroup',
            name='intro',
            field=models.TextField(default=''),
        ),
        migrations.AddField(
            model_name='project',
            name='intro',
            field=models.TextField(default=''),
        ),
    ]



# Generated by Django 3.0.3 on 2020-04-29 13:47

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('experiment', '0008_variable_default_val'),
    ]

    operations = [
        migrations.AlterField(
            model_name='singlevalue',
            name='value',
            field=models.CharField(max_length=100),
        ),
        migrations.AlterField(
            model_name='variable',
            name='default_val',
            field=models.CharField(default='', max_length=100),
        ),
    ]



"""
Django settings for fitterlog_server project.

Generated by 'django-admin startproject' using Django 3.0.3.

For more information on this file, see
https://docs.djangoproject.com/en/3.0/topics/settings/

For the full list of settings and their values, see
https://docs.djangoproject.com/en/3.0/ref/settings/
"""

import os

# Build paths inside the project like this: os.path.join(BASE_DIR, ...)
BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))


# Quick-start development settings - unsuitable for production
# See https://docs.djangoproject.com/en/3.0/howto/deployment/checklist/

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = 'p@03qa2o@anbi^z)mx#+efk3c&bbs9ryh#+6cs^be^*+reo%tc'

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = True

ALLOWED_HOSTS = ["*"]


# Application definition
INSTALLED_APPS = [
	'django.contrib.admin',
	'django.contrib.auth',
	'django.contrib.contenttypes',
	'django.contrib.sessions',
	'django.contrib.messages',
	'django.contrib.staticfiles',
	'experiment.apps.ExperimentConfig',
]

MIDDLEWARE = [
	'django.middleware.security.SecurityMiddleware',
	'django.contrib.sessions.middleware.SessionMiddleware',
	'django.middleware.common.CommonMiddleware',
	'django.middleware.csrf.CsrfViewMiddleware',
	'django.contrib.auth.middleware.AuthenticationMiddleware',
	'django.contrib.messages.middleware.MessageMiddleware',
	'django.middleware.clickjacking.XFrameOptionsMiddleware',
]

ROOT_URLCONF = 'fitterlog_server.urls'

TEMPLATES = [
	{
		'BACKEND': 'django.template.backends.django.DjangoTemplates',
		'DIRS': [],
		'APP_DIRS': True,
		'OPTIONS': {
			'context_processors': [
				'django.template.context_processors.debug',
				'django.template.context_processors.request',
				'django.contrib.auth.context_processors.auth',
				'django.contrib.messages.context_processors.messages',
			],
		},
	},
]

WSGI_APPLICATION = 'fitterlog_server.wsgi.application'


# Database
# https://docs.djangoproject.com/en/3.0/ref/settings/#databases

DATABASES = {
	'default': {
		'ENGINE': 'django.db.backends.sqlite3',
		'NAME': os.path.join(BASE_DIR, 'db.sqlite3'),
		'OPTIONS': {
			'timeout': 60,
		}
	}
}


# Password validation
# https://docs.djangoproject.com/en/3.0/ref/settings/#auth-password-validators

AUTH_PASSWORD_VALIDATORS = [
	{
		'NAME': 'django.contrib.auth.password_validation.UserAttributeSimilarityValidator',
	},
	{
		'NAME': 'django.contrib.auth.password_validation.MinimumLengthValidator',
	},
	{
		'NAME': 'django.contrib.auth.password_validation.CommonPasswordValidator',
	},
	{
		'NAME': 'django.contrib.auth.password_validation.NumericPasswordValidator',
	},
]


# Internationalization
# https://docs.djangoproject.com/en/3.0/topics/i18n/

LANGUAGE_CODE = 'en-us'

TIME_ZONE = 'Asia/Shanghai'

USE_I18N = True

USE_L10N = True

USE_TZ = True


# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/3.0/howto/static-files/

STATIC_URL = '/static/'



# Generated by Django 3.0.3 on 2020-04-26 07:53

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('experiment', '0003_auto_20200426_1535'),
    ]

    operations = [
        migrations.AddField(
            model_name='experiment',
            name='start_time',
            field=models.DateField(auto_now=True),
        ),
    ]



class FitterType:

	def __init__(self , func , name):
		self.name = name
		self.func = func

	def __call__(self , *pargs , **kwargs):

		return self.func(*pargs , **kwargs)

	def __str__(self):
		return self.name


Int 	= FitterType(int 												, "Int"   )
Float 	= FitterType(lambda x: round(float(x) , 6) 						, "Float" )
String 	= FitterType(str 												, "String")
Bool 	= FitterType(lambda x: x == "True" or x == "1" or x == "true" 	, "Bool"  )

# 特殊类型，表示在前端放一根分隔线
FITTER_SPLITTER = FitterType(lambda x: print("You should not see this."), "_FITTER_SPLITTER")